import contextlib
import logging
import re
from typing import Union, Callable, List, Optional

import requests
from pypsrp.client import Client
from pypsrp.exceptions import WinRMError
from pypsrp.powershell import PSDataStreams

from . import exceptions

logger = logging.getLogger(__name__)
logging.getLogger("urllib3").setLevel(logging.ERROR)
logging.getLogger("pypsrp").setLevel(logging.ERROR)


def _get_session(
    *, host: str, port: Union[str, int], username: str, password: str, validate_cert: bool, use_ssl: bool
) -> Callable:
    @contextlib.contextmanager
    def _open():
        try:
            logger.info(f"opening psrp client to {host}:{port} as {username} with ssl: {use_ssl}")
            psrp_client = Client(
                host,
                username=username,
                password=password,
                cert_validation=validate_cert,
                ssl=use_ssl,
                port=port,
                encryption="auto",
            )
            # Getting the server config tests the connection for us.
            psrp_client.wsman.get_server_config()
            yield psrp_client

        except requests.exceptions.SSLError as e:
            logger.info(e)
            raise exceptions.CertificateError(
                "Unable to verify SSL certificate. If the certificate is "
                "self-signed, pass validate_cert=False to "
                "windows_server.client.initialize()."
            )

        except requests.exceptions.ConnectionError as e:
            match = re.compile(r".*(\[Errno \d+\]\s[\w\s]+)").search(str(e))
            error: Union[None, str] = None
            if match:
                error = match.group(1)
            raise exceptions.NetworkError(
                f"Connection error sending {e.request.method} " f"request to {e.request.url}. Error: {error}."
            )

        finally:
            # No need to close the psrp_client.
            pass

    return _open


def _run_dos(connection: Client, *, command: str, environment: Optional[dict] = None) -> str:
    logger.info("executing DOS command")
    std_out, std_err, _ = connection.execute_cmd(command, environment=environment)
    if std_err:
        raise exceptions.DOSCommandError(std_err)
    return std_out


def _collect_errors(streams: PSDataStreams) -> List[str]:
    """ If a powershell script makes a connection to a
    different server, for example using Invoke-Command,
    that sub-script may fail.

    If it does fail, the execute_ps method does not set
    the failed flag, so if the error stream isn't empty
    we have to assume failure and inspect all the streams.
    """
    logger.debug("there are error streams")
    errors = []
    for stream_collection in (stream for stream in (streams.error, streams.warning, streams.debug)):
        for s in stream_collection:
            if str(s) != s.message:
                errors.append(f"{str(s)}: {s.message}")
            else:
                errors.append(str(s))

    errors = list(set(errors))
    logger.debug(f"collected errors: {errors}")
    return errors


def _run_powershell(
    connection: Client,
    *,
    script: str,
    configuration_name: str = "Microsoft.PowerShell",
    environment: Optional[dict] = None,
) -> str:
    logger.info(f"executing powershell script with configuration_name {configuration_name}")
    std_out, streams, _ = connection.execute_ps(script, configuration_name=configuration_name, environment=environment)
    if streams.error:
        errors = _collect_errors(streams)
        if errors:
            raise exceptions.PowershellError(", ".join(errors))
        else:
            logger.debug(f"no error messages were found in available streams {str(streams)}")
            logger.debug(f"stdout: {std_out}")
            raise exceptions.PowershellError(
                "Script failed but no errors were found in the "
                "errors, warning, or debug streams. Please view "
                "the logs for more information."
            )
    else:
        return std_out


def _upload_file(
    connection: Client, *, source: str, destination: str, configuration_name: str = "Microsoft.PowerShell"
) -> str:
    try:
        return connection.copy(source, destination, configuration_name=configuration_name)
    except WinRMError as e:
        logger.error(e)
        winrm_error_string = "".join(e.args).replace("\n", "").replace("\r", "")
        if "DirectoryNotFoundException" in winrm_error_string:
            raise exceptions.PathError(f"Supplied destination path is invalid: '{destination}'")
        elif "being used by another process" in winrm_error_string:
            raise exceptions.CopyFileError(
                "Failed to upload file. The file is being used " "by another process on the server."
            )
        raise exceptions.CopyFileError("Failed to copy file to server. Please check the logs for more information.")
    except FileNotFoundError as e:
        logger.error(str(e))
        raise exceptions.PathError(f"Supplied source path does not exist: '{source}'")


def _download_file(
    connection: Client, *, source: str, destination: str, configuration_name: str = "Microsoft.PowerShell"
):
    try:
        return connection.fetch(source, destination, configuration_name=configuration_name)
    except WinRMError as e:
        logger.error(e)
        winrm_error_string = "".join(e.args).replace("\n", "").replace("\r", "")
        if re.search(re.compile(r"The\spath\sat\s.*\sdoes\snot\sexist"), repr(winrm_error_string)):
            raise exceptions.PathError(f"Supplied source path is invalid: '{source}'")
        raise exceptions.CopyFileError("Failed to copy file to server. Please check the logs for more information.")
    except FileNotFoundError as e:
        logger.error(str(e))
        raise exceptions.PathError(f"Supplied destination path does not exist: '{destination}'")
