from typing import TypedDict

from datacaster.classes import CastDataClass


RDP_SESSION_STATES = {"disc": "Disconnected", "listen": "Listening", "conn": "Connected", "active": "Active"}


class WindowsServer(TypedDict):
    r"""
- Contains remote Powershell connection details for a Windows Server

- It is used as an _input_ to many of the actions in this plugin

- The example below shows the dictionary structure in YAML

- See the ___required___ and _optional_ keys underneath the example

!!! warning "Supplying Credentials"

    You should ___always___ use a PPA Vault integration to provide credentials to a plugin action.

!!! example

    ```yaml linenums="1"
    windows_server:
      address: windows-server.internal.net
      username: example.username
      password: example.password
      validate_cert: true
      use_ssl: true
    ```

???+ danger "Required Keys"

    __address__: The target IP or DNS address.

    __username__: Username for authentication.

    __password__: Password for authentication.

    __validate_cert__: Set to _false_ to skip certificate validation (defaults to _true_).

    __use_ssl__: Set to _false_ to connect over HTTP (defaults to _true_ using HTTPS).
    """
    address: str
    username: str
    password: str
    validate_cert: bool
    use_ssl: bool


class RDPSessionData(CastDataClass):
    id: int
    username: str
    session_name: str
    state: str
    __class_config__ = {
        "rename_fields": {"sessionname": "session_name"},
        "cast_functions": {"fields": {"state": lambda state: RDP_SESSION_STATES.get(state.lower(), state)}},
        "always_cast": ["state"],
    }


class RDPSession(TypedDict):
    r"""
- Contains RDP session details

- The example below shows the dictionary structure in YAML

!!! example

    ```yaml linenums="1"
    rdp_session:
      username: domain\username
      session_name: rdp-tcp#50
      state: Active
      id: 2
    ```

???+ info "All RDPSession Keys"

    __id__: The RDP session ID on the Windows Server.

    __username__: The username the RDP session was authenticated with.

    __session_name__: The name of the RDP session.

    __state__: The state of the RDP session (`Disconnected`, `Connected`, `Active`, `Listening`).
    """
    __annotations__ = RDPSessionData.__annotations__


class ServiceData(CastDataClass):
    name: str
    display_name: str
    state: str
    status: str
    start_name: str
    description: str
    __class_config__ = {
        "rename_fields": {"displayname": "display_name", "startname": "start_name"},
    }


class Service(TypedDict):
    r"""
- Contains Windows service details

- The example below shows the dictionary structure in YAML

!!! example

    ```yaml linenums="1"
    name: Dhcp
    display_name: DHCP Client
    state: Running
    status: OK
    start_name: NT Authority\\LocalService
    description: >
      Registers and updates IP addresses and DNS records for this computer.
      If this service is stopped, this computer will not receive dynamic IP addresses
      and DNS updates. If this service is disabled, any services that explicitly depend
      on it will fail to start.
    ```

???+ info "All Service Keys"

    __name__: The name of the service.

    __display_name__: The display name for the service.

    __state__: The state of the service.

    __status__: The service status.

    __start_name__: The account the service runs under.

    __description__: The service description.
    """
    __annotations__ = ServiceData.__annotations__
