"""
Services

This module contains actions for viewing & managing services on a Windows Server.
"""
import uuid

from typing import List, Optional

from ppa import ui

from windows_server import services

from .records import WindowsServer, Service, ServiceData
from ._wrappers import winrm


SERVICE_TABLE_HEADER = ["Display Name", "Name", "State", "Status"]

SERVICE_TABLE_FIELDS = [
    lambda service: service.get("display_name") or "",
    lambda service: service.get("name"),
    lambda service: service.get("state"),
    lambda service: service.get("status"),
]


class NoServicesFound(Exception):
    pass


def _create_services(raw_sessions):
    return [ServiceData(**session._asdict()).__dict__ for session in raw_sessions]


@winrm
def get_all(windows_server: WindowsServer) -> List[Service]:
    r"""Get all services.

!!! question "Minimum Plugin Version: 1.2.0"

??? info "Input"

    - __windows_server:__ a [WindowsServer](../inputs_outputs/#windowsserver) dictionary

??? success "Output"

    A list of [Service](../inputs_outputs/#service) dictionaries.

??? example

    ```YAML linenums="1"
    - windows_server.services.get_all:
      load:
        windows_server: windows_server
      save: all_services
    ```
    """
    return _create_services(services.get_all(windows_server))


@winrm
def get_by_name(windows_server: WindowsServer, name: str) -> Service:
    r"""Get a service using its name.

!!! question "Minimum Plugin Version: 1.2.0"

??? info "Input"

    - __windows_server:__ a [WindowsServer](../inputs_outputs/#windowsserver) dictionary

    - __name:__ the name of the service

??? success "Output"

    A single [Service](../inputs_outputs/#service) dictionary.

??? example

    ```YAML linenums="1"
    - windows_server.services.get_by_name:
        name: spooler
      load:
        windows_server: windows_server
      save: spooler_service
    ```
    """
    try:
        return _create_services(services.get_by_name(windows_server, name=name))[0]
    except IndexError:
        raise NoServicesFound(f"No service was found with name '{name}'.")


@winrm
def restart(windows_server: WindowsServer, name: str) -> None:
    r"""Restart a service using its name.

!!! question "Minimum Plugin Version: 1.2.0"

??? info "Input"

    - __windows_server:__ a [WindowsServer](../inputs_outputs/#windowsserver) dictionary

    - __name:__ the name of the service to restart

??? success "Output"

    Nothing is outputted by this action.

??? example

    ```YAML linenums="1"
    - windows_server.services.restart:
        name: spooler
      load:
        windows_server: windows_server
    ```
    """
    services.restart(windows_server, name=name)


@winrm
def start(windows_server: WindowsServer, name: str) -> None:
    r"""Start a service using its name.

!!! question "Minimum Plugin Version: 1.2.0"

??? info "Input"

    - __windows_server:__ a [WindowsServer](../inputs_outputs/#windowsserver) dictionary

    - __name:__ the name of the service to start

??? success "Output"

    Nothing is outputted by this action.

??? example

    ```YAML linenums="1"
    - windows_server.services.start:
        name: spooler
      load:
        windows_server: windows_server
    ```
    """
    services.start(windows_server, name=name)


@winrm
def stop(windows_server: WindowsServer, name: str) -> None:
    r"""Stop a service using its name.

!!! question "Minimum Plugin Version: 1.2.0"

??? info "Input"

    - __windows_server:__ a [WindowsServer](../inputs_outputs/#windowsserver) dictionary

    - __name:__ the name of the service to stop

??? success "Output"

    Nothing is outputted by this action.

??? example

    ```YAML linenums="1"
    - windows_server.services.stop:
        name: spooler
      load:
        windows_server: windows_server
    ```
    """
    services.stop(windows_server, name=name)


def input_table(
    text: str, service_list: List[Service], minimum: Optional[int] = None, maximum: Optional[int] = None
) -> List[Service]:
    """Display a list of Windows services in a table, & allow the task operator to make a selection.

The table will have the following columns:

- Display Name
- Name
- State
- Status

!!! question "Minimum Plugin Version: 1.1.0"

??? info "Input"

    - __text:__ the title of the table

    - __service_list:__ a list of [Service](../inputs_outputs/#service) dictionaries to display in the table

    - __minimum:__ the minimum number of acceptable selections

    - __maximum:__ the maximum number of acceptable selections

??? success "Output"

    A list of [Service](../inputs_outputs/#service) dictionaries.

??? tip

    - If neither a _minimum_ or _maximum_ is provided, the task operator will be able to submit 0 selections.

    - If _minimum_ or _maximum_ are provided, the operation will repeat until the task operator makes a valid number
        of selections.

??? example

    - Getting all services & saving them as `all_services`

    - Using this action to display them in a table, & saving the selection as `selected_service`

    ```YAML linenums="1" hl_lines="4 11 13"
    - windows_server.services.get_all:
      load:
        windows_server: windows_server
      save: all_services

    - windows_server.services.input_table:
        text: Please Select a Session
        minimum: 1
        maximum: 1
      load:
        session_list: all_services
      save: selected_service
    ```
    """

    table_element_id = str(uuid.uuid4())
    error_element_id = str(uuid.uuid4())
    while True:
        indices = ui.input_table(
            text,
            header=SERVICE_TABLE_HEADER,
            data=[[get(service) for get in SERVICE_TABLE_FIELDS] for service in service_list],
            element_id=table_element_id,
        )
        count = len(indices)
        if minimum and count < minimum:
            ui.output_error(f"Please choose at least {minimum} record(s)", element_id=error_element_id)
            continue
        if maximum and count > maximum:
            ui.output_error(f"Please choose no more than {maximum} record(s)", element_id=error_element_id)
            continue
        break

    return [service_list[i] for i in indices]


def output_table(text: str, services: List[Service]) -> None:
    """Display one or more Windows services in a table.

The table will have the following columns:

- Display Name
- Name
- State
- Status

!!! question "Minimum Plugin Version: 1.2.0"

??? info "Input"

    - __text:__ the title of the table

    - __services:__ a single or list of [Service](../inputs_outputs/#service) dictionaries to display in the table

??? success "Output"

    Nothing is outputted by this action.

??? example

    - Getting all services & saving them as `all_services`

    - Using this action to display them in a table

    ```YAML linenums="1" hl_lines="4 9"
    - windows_server.services.get_all:
      load:
        windows_server: windows_server
      save: all_services

    - windows_server.services.output_table:
        text: All Services
      load:
        services: all_services
    ```
    """
    if not isinstance(services, list):
        services = [services]
    ui.output_table(
        text, header=SERVICE_TABLE_HEADER, data=[[get(service) for get in SERVICE_TABLE_FIELDS] for service in services],
    )
